@@ -1,5 +1,7 @@
 /*****************************************************************************\
- *  $Id: hostlist.c 5700 2005-08-19 20:45:00Z grondo $
+ *  $Id: hostlist.c 11809 2007-07-10 23:10:05Z grondo $
+ *****************************************************************************
+ *  $LSDId: hostlist.c 2955 2006-07-27 14:48:33Z grondo $
  *****************************************************************************
  *  Copyright (C) 2002 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -97,6 +99,9 @@
 /* max host range: anything larger will be assumed to be an error */
 #define MAX_RANGE    16384    /* 16K Hosts */
 
+/* max host suffix value */
+#define MAX_HOST_SUFFIX 1<<25
+
 /* max number of ranges that will be processed between brackets */
 #define MAX_RANGES    10240    /* 10K Ranges */
 
@@ -347,6 +352,22 @@ static void _error(char *file, int line, char *msg, ...)
     return;
 }
 
+static int _advance_past_brackets (char *tok, char **str)
+{
+    /* if _single_ opening bracket exists b/w tok and str, push str
+     * past first closing bracket to next seperator */
+    if (   memchr(tok, '[', *str - tok) != NULL
+        && memchr(tok, ']', *str - tok) == NULL ) {
+        char *q = strchr(*str, ']');
+        if (q && memchr(*str, '[', q - *str) == NULL) {
+            *str = q + 1;
+            return (1);
+        }
+    }
+
+    return 0;
+}
+
 /* 
  * Helper function for host list string parsing routines 
  * Returns a pointer to the next token; additionally advance *str
@@ -371,20 +392,18 @@ static char * _next_tok(char *sep, char **str)
     /* assign token ptr */
     tok = *str;
 
-    /* push str past token and leave pointing to first separator */
-    while (**str != '\0' && strchr(sep, **str) == '\0')
-        (*str)++;
+    /*
+     * Advance str past any separators, but if a separator occurs between
+     *  brackets, e.g. foo[0-3,5], then advance str past closing brackets and
+     *  try again.
+     */
+    do {
+        /* push str past token and leave pointing to first separator */
+        while (**str != '\0' && strchr(sep, **str) == '\0')
+            (*str)++;
+    } while (_advance_past_brackets (tok, str));
 
-    /* if _single_ opening bracket exists b/w tok and str, push str
-     * past first closing bracket */
-    if (   memchr(tok, '[', *str - tok) != NULL
-        && memchr(tok, ']', *str - tok) == NULL ) {
-        char *q = strchr(*str, ']');
-        if (q && memchr(*str, '[', q - *str) == NULL)
-            *str = q + 1;
-    }
-
-    /* nullify consecutive separators and push str beyond them */
+   /* nullify consecutive separators and push str beyond them */
     while (**str != '\0' && strchr(sep, **str) != '\0')
         *(*str)++ = '\0';
 
@@ -498,7 +517,7 @@ static hostname_t hostname_create(const char *hostname)
     hn->suffix = hn->hostname + idx + 1;
     hn->num = strtoul(hn->suffix, &p, 10);
 
-    if (*p == '\0') {
+    if ((*p == '\0') && (hn->num <= MAX_HOST_SUFFIX)) {
         if (!(hn->prefix = malloc((idx + 2) * sizeof(char)))) {
             hostname_destroy(hn);
             out_of_memory("hostname prefix create");
@@ -891,6 +910,9 @@ static int hostrange_hn_within(hostrange_t hr, hostname_t hn)
 {
     int retval = 0;
 
+    if (hr->singlehost && (strcmp(hn->hostname, hr->prefix) == 0))
+        return 1;
+
     if (strcmp(hr->prefix, hn->prefix) == 0) {
         if (!hostname_suffix_is_valid(hn)) {
             if (hr->singlehost)
@@ -1388,6 +1410,28 @@ _push_range_list(hostlist_t hl, char *pfx, struct _range *rng,
     }
 }
 
+static void
+_push_range_list_with_suffix(hostlist_t hl, char *pfx, char *sfx, 
+                             struct _range *rng, int n)
+{
+    int i;
+    unsigned long j;
+    for (i = 0; i < n; i++) {
+        for (j = rng->lo; j <= rng->hi; j++) {
+            char host[4096];
+            hostrange_t hr;
+            snprintf (host, 4096, "%s%0*lu%s", pfx, rng->width, j, sfx);
+            hr = hostrange_create_single (host);
+            hostlist_push_range (hl, hr);
+            /*
+             * hr is copied in hostlist_push_range. Need to free here.
+             */
+            hostrange_destroy (hr);
+        }
+        rng++;
+    }
+}
+
 /*
  * Create a hostlist from a string with brackets '[' ']' to aid 
  * detection of ranges and compressed lists
@@ -1421,7 +1465,11 @@ _hostlist_create_bracketed(const char *hostlist, char *sep, char *r_op)
                 nr = _parse_range_list(p, ranges, MAX_RANGES);
                 if (nr < 0) 
                     goto error;
-                _push_range_list(new, prefix, ranges, nr);
+
+                if (*(++q) != '\0')
+                    _push_range_list_with_suffix (new, prefix, q, ranges, nr);
+                else
+                    _push_range_list(new, prefix, ranges, nr);
 
                 
             } else
@@ -1818,7 +1866,7 @@ int hostlist_find(hostlist_t hl, const char *hostname)
 
     for (i = 0, count = 0; i < hl->nranges; i++) {
         if (hostrange_hn_within(hl->hr[i], hn)) {
-            if (hostname_suffix_is_valid(hn))
+            if (hostname_suffix_is_valid(hn) && !hl->hr[i]->singlehost)
                 ret = count + hn->num - hl->hr[i]->lo;
             else
                 ret = count;
@@ -2201,7 +2249,8 @@ static void _iterator_advance_range(hostlist_iterator_t i)
 
 char *hostlist_next(hostlist_iterator_t i)
 {
-    char buf[MAXHOSTNAMELEN + 16];
+    char *buf = NULL;
+    char suffix[16];
     int len = 0;
     assert(i != NULL);
     assert(i->magic == HOSTLIST_MAGIC);
@@ -2213,12 +2262,21 @@ char *hostlist_next(hostlist_iterator_t i)
         return NULL;
     }
 
-    len = snprintf(buf, MAXHOSTNAMELEN + 15, "%s", i->hr->prefix);
+    suffix[0] = '\0';
+
     if (!i->hr->singlehost)
-        snprintf(buf + len, MAXHOSTNAMELEN + 15 - len, "%0*lu",
-             i->hr->width, i->hr->lo + i->depth);
+        snprintf (suffix, 15, "%0*lu", i->hr->width, i->hr->lo + i->depth);
+
+    len = strlen (i->hr->prefix) + strlen (suffix) + 1;
+    if (!(buf = malloc (len)))
+        out_of_memory("hostlist_next");
+    
+    buf[0] = '\0';
+    strcat (buf, i->hr->prefix);
+    strcat (buf, suffix);
+
     UNLOCK_HOSTLIST(i->hl);
-    return strdup(buf);
+    return (buf);
 }
 
 char *hostlist_next_range(hostlist_iterator_t i)
